/* 
    SIT - SIS is tiny, intended to be a IRC Client for secure communication
	on top of the IRC protocol as described in RFC 1459
    Copyright (C) 2000  Christian Leber

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    e-mail: c_l@bigfoot.com
    I hope this is GPL compliant, if not please contact me.
*/

#include <stdlib.h>
#include <stdio.h>
#include <gmp.h>

#include "random.h"

int power(int base,int exp)
{
	int i,p=1;
	for(i=1;i<=exp;++i) p = p * base;
	return p;
}


void big_pprime(mpz_t output, int length, int shouldprime)
{
	unsigned long int *random;
	unsigned char *random_inp;
	unsigned long int random_z;
	int i;
	unsigned int random_bit; /* "points" to the first bit witch is needed, we have to set it to 1,
	                   this can be from 7 to 0 */	                
	div_t erg;
	mpz_t addi_mpz, mult_mpz, rand_mpz, r_mpz;

	erg=div(length,8);
	
	/* let's have a look how many random bytes we need... */
	if(erg.rem==0) {random_z = erg.quot; random_bit=7;}
		else {random_z = erg.quot + 1; random_bit = erg.rem-1;}

	random = (unsigned long int *)malloc(random_z*4);
	
	printf("%s\n%s\n","  we're going to get some random data... this may take time, you",
			"  should make some activities(disk, network and so)");

	random_inp=n_byte_random(random_z);

/*	read(fileno(devrand),random_inp,random_z);*/

/*	for(i=0;i<random_z;i++)	random_inp[i] = getc(devrand);*/
	for(i=0;i<random_z;i++)	random[i] = random_inp[i];
	
	
	/* ok, in generell we set the bits with the highest values, that are
	out of out bitrange to 0 */
	/* if shouldprime=1 we set the bit with the lowest value to 1 and the
	bit with the highest value in our range to 1 */
	if(shouldprime==1)
	{
		if(random[0] % 2 == 0) random[0]+=1;
		for(i=7;i>=random_bit;i--)
		{
			if( (random[random_z-1]/power(2,i)) == 1)
				{random[random_z-1] = random[random_z-1]-power(2,i);}
		}
		random[random_z-1]=random[random_z-1]+power(2,random_bit);
	}
	else
	{
		for(i=7;i>random_bit;i--)
		{
			if( (random[random_z-1]/power(2,i)) == 1)
				{random[random_z-1] = random[random_z-1]-power(2,i);}
		}
	}

/*	printf("now: %lu\n",random[random_z-1]);
	
	printf("random_z: %lu random_bit: %d",random_z,random_bit);
	printf("\nrandomstuff: ");
	for(i=random_z-1;i>=0;i--)
	{	
		printf("%u ",random_inp[i]);
		printf("%lu ",random[i]);
	}
	printf("\n");*/
			
	mpz_init(mult_mpz);
	mpz_init(addi_mpz);
	mpz_init(rand_mpz);
	mpz_init(r_mpz);
	mpz_set_ui(rand_mpz,random[0]);
	mpz_set_ui(mult_mpz,1);

	for(i=1;i<random_z;i++)
	{
		mpz_mul_ui(mult_mpz,mult_mpz,256);
		mpz_set_ui(r_mpz,random[i]);
		mpz_mul(addi_mpz,mult_mpz,r_mpz);
		mpz_add(rand_mpz,rand_mpz,addi_mpz);
	}
	
	free(random);
	mpz_set(output,rand_mpz);
	mpz_clear(mult_mpz);
	mpz_clear(addi_mpz);
	mpz_clear(rand_mpz);
	mpz_clear(r_mpz);

/*	mpz_out_str(stdout, 10, output);
	printf("\n");*/
}

void clrsa_probab_prime(mpz_t output,int bit_length)
{
	mpz_t pprime;
	
	mpz_init(pprime);
	big_pprime(pprime,bit_length,1);

	while(mpz_probab_prime_p(pprime,10)==0)
		mpz_add_ui(pprime,pprime,2);
		
	mpz_set(output,pprime);
	mpz_clear(pprime);
}

