/*
    SIT - SIS is tiny, intended to be a IRC Client for secure communication
	on top of the IRC protocol as described in RFC 1459
    Copyright (C) 2000  Christian Leber

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    e-mail: c_l@bigfoot.com
    I hope this is GPL compliant, if not please contact me.
*/


#include <stdio.h>
#include <stdlib.h>
#include <gmp.h>

#include "all.h"
#include "databox.h"

DataBox *DataBox_init(int size)
{
	DataBox *output;
	
	output = (DataBox *)malloc(sizeof(DataBox));
	if(size>0)
	{
		output->data = (unsigned char *)malloc(size);
		/* FIXME: we should check the return value */
		output->reserved=size;
	}
	output->init=true;
	return output;	
}


DataBox *DataBox_mpz_to_DataBox(mpz_t input)
{
	unsigned long length2, length8, i;
	mpz_t rem, back;
	div_t erg;		
	DataBox *output;

	/* proof if input > 0 */
	if(mpz_cmp_ui(input,0)<=0) return NULL;

	length2=mpz_sizeinbase(input,2);

	erg=div(length2,8);
	if(erg.rem==0) length8 = erg.quot;
		else length8 = erg.quot + 1;
		
	output = DataBox_init(length8);
	output->length = length8;
	
	mpz_init(rem);
	mpz_init_set(back,input);

    for (i=0;i<length8;i++)
	{
		mpz_tdiv_qr_ui(back, rem, back, 256);
		output->data[i]=mpz_get_ui(rem);
    }

	mpz_clear(back);
	mpz_clear(rem);
	
	return output;
}


bool DataBox_DataBox_to_mpz(mpz_t output, DataBox *input)
{
	int i;
	mpz_init(output);
	
	if(input==NULL) return false;
	if(input->length < 1) return false;

    for(i=input->length;i>=0;i--)	{
		mpz_mul_2exp (output, output, 8);
		mpz_add_ui (output, output, input->data[i]);
    }
	return true;
}


char *DataBox_to_base64(DataBox input)
{
	char dtable[64] = {65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,
		82,83,84,85,86,87,88,89,90,97,98,99,100,101,102,103,104,105,106,107,
		108,109,110,111,112,113,114,115,116,117,118,119,120,121,122,48,49,
		50,51,52,53,54,55,56,57,43,47};
	/* this is needed for base64,
					 FIXME: perhaps a statically table would be faster*/
	int i,length3;
	int inpos; /* actual position in the input string inside the Databox */
	int outpos=0; /* actual position in the output string */
	unsigned char *buff_2;
	char *output;
	
	if(input.length==0) return NULL;
	length3 = (input.length / 3);
	if( (input.length % 3) != 0 ) length3 = length3 + 1;

	output = (char *)malloc( (length3 * 4) + 1);

	buff_2 = (unsigned char *)malloc( (length3 * 3) + 1);
	for(i=0;i<input.length;i++)	{buff_2[i] = input.data[i];}
	for(i=input.length;i<length3*3;i++) {buff_2[i] = 0;} 
	
    /*	Fill up dtable with the neccessary character encodings.  */

/*	for (i = 0; i < 26; i++)
		{dtable[i] = 'A' + i; dtable[26 + i] = 'a' + i;}
	for (i = 0; i < 10; i++)
		{dtable[52 + i] = '0' + i;}
    dtable[62] = '+'; dtable[63] = '/';*/

	for(i=0;i<length3;i++)
	{
		outpos = 4 * i;
		inpos = 3 * i;
		output[outpos] = dtable[buff_2[inpos] >> 2];
		output[outpos + 1] = dtable[((buff_2[inpos] & 3) << 4) | (buff_2[inpos + 1] >> 4)];
		output[outpos + 2] = dtable[((buff_2[inpos + 1] & 0xF) << 2) | (buff_2[inpos + 2] >> 6)];
		output[outpos + 3] = dtable[buff_2[inpos + 2] & 0x3F];
	}

	if (input.length < length3*3)
		{
			output[outpos + 3] = '=';
			if ( (input.length + 1) < length3*3) {output[outpos + 2] = '=';}
	    }

	output[length3*4] = 0;

	return output;
}


DataBox base64_to_DataBox(char *input)
{
	DataBox output;
	int c,i,j,length4,pos=0;
	int length_input; /* the length of *input */
	unsigned char dtable[256]={128,128,128,128,128,128,128,128,128,128,128,
		128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
		128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,62,128,
		128,128,63,52,53,54,55,56,57,58,59,60,61,128,128,128,0,128,128,128,
		0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,
		25,128,128,128,128,128,128,26,27,28,29,30,31,32,33,34,35,36,37,38,
		39,40,41,42,43,44,45,46,47,48,49,50,51,128,128,128,128,128,128,128,
		128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
		128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
		128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
		128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
		128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
		128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
		128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
		128,128,128,128,128,128,128};
	
	unsigned char a[4], b[4];

	length_input = strlen(input);

	if(length_input == 0){output.data=NULL;return output;}
	length4 = length_input / 4;
	if(length4 * 4 != length_input) {output.data=NULL;return output;}

	output.data = (unsigned char *)malloc(length4*3);
	output.length = length4 * 3;

	/* Create the for base64 neccessary table */
/*	for (i = 0; i < 256; i++)
		{ dtable[i] = 0x80; }
	for (i = 'A'; i <= 'Z'; i++)
		{ dtable[i] = 0 + (i - 'A'); }
    for (i = 'a'; i <= 'z'; i++)
		{ dtable[i] = 26 + (i - 'a'); }
    for (i = '0'; i <= '9'; i++)
		{ dtable[i] = 52 + (i - '0'); }
	dtable['+'] = 62;
	dtable['/'] = 63;
	dtable['='] = 0;*/

/*	for(i=0;i<256;i++)printf("%u,",dtable[i]);printf("\n");*/

	for(j=0;j<length4;j++)
	{
		for (i = 0; i < 4; i++)
		{
			c = input[(j*4) + i];
 		    if (dtable[c] & 0x80)
		    {
				printf("%s\n","invalid character, exiting base64_to_DataBox");
				output.data=NULL;
				return output;
 		/* Ignoring errors: discard invalid character. */
			/*	i--;
				continue;*/
			}
			a[i] = c;
			b[i] = dtable[c];
		}
	
		output.data[pos++] = (b[0] << 2) | (b[1] >> 4);
		output.data[pos++] = (b[1] << 4) | (b[2] >> 2);
		output.data[pos++] = (b[2] << 6) | b[3];
		i = a[2] == '=' ? 1 : (a[3] == '=' ? 2 : 3);
	}
	
	if(a[3]=='=') output.length = output.length - 1;
	if(a[2]=='=') output.length = output.length - 1;
	
	return output;
}

#ifdef ALONE
int main()
{
	mpz_t test01_mpz;
	DataBox *test02=NULL;
	mpz_t test03_mpz;

	mpz_init_set_str(test01_mpz,"234623742375423637456345745672346688889798056735612341842638",10);
	mpz_out_str(stdout, 10, test01_mpz);
	printf("\n");
	test02 = DataBox_mpz_to_DataBox(test01_mpz);
	DataBox_DataBox_to_mpz(test03_mpz,test02);	
	mpz_out_str(stdout, 10, test03_mpz);
	printf("\n");
}

#endif
