/* 
    SIT - SIS is tiny, intended to be a IRC Client for secure communication
	on top of the IRC protocol as described in RFC 1459
    Copyright (C) 2000  Christian Leber

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    e-mail: c_l@bigfoot.com
    I hope this is GPL compliant, if not please contact me.
*/

#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <gmp.h>
#include "dstruct.h"
#include "../blowfish/common.h"
#include "../blowfish/blowfish.h"

/*#define ALONE*/

#ifdef ALONE
static struct node_channel_list *DS_channel;
static struct node_nick_list *DS_nick[DS_NICK_HASHSIZE];
static struct node_s_list *DS_s;
#endif

/* ######################################################## */
/* for the BLOWFISH                                         */

struct DS_blowfish *DS_blowfish_init()
{
	struct DS_blowfish *output;

	output = (struct DS_blowfish *)malloc(sizeof(*output));
	output->blowfish_key=NULL;
	output->blowfish_length=0;
/*	output->blowfish_schedule=NULL;*/
	output->blowfish_init=false;
	
	return output;
}

void DS_blowfish_schedule(struct DS_blowfish *input)
{
	if(input->blowfish_length>72) return;
/*	input->blowfish_sched = (WORD32 *)malloc(1043);*/
	blowfish_key_schedule(input->blowfish_key,input->blowfish_length,input->blowfish_sched);
	input->blowfish_init=true;
}


/* ######################################################## */
/* for the NICKS                                            */


/* this returns a hash value for the given string, the
   string should have been set tolower allready...          */      
unsigned DS_nick_hash(char *s)
{
	unsigned hashval;
	
	for(hashval = 0; *s != '\0'; s++)
		hashval = *s + 31 * hashval;
	return hashval % DS_NICK_HASHSIZE;
}

/* this returns a pointer to the data structure of the      */
/* given nick                                               */
struct node_nick_list *DS_nick_lookup(struct node_s_list *this_DS_s, char *nick)
{
	struct node_nick_list *np;
	char *nick_tolower;
	nick_tolower = irc_tolower(nick);
	
	for(np=this_DS_s->DS_nick[DS_nick_hash(nick_tolower)]; np != NULL; np = np->next)
		if(strcmp(nick_tolower, np->nick) == 0)
		{
			free(nick_tolower);
			return np; /* do issa ja */
		}
	free(nick_tolower);
	return NULL; /* verdammt */
}

struct node_nick_list *DS_nick_install(struct node_s_list *this_DS_s, char *onick)
{
	struct node_nick_list *np;
	unsigned hashval;
	char *nick; /* set tolower, we point to this so we can't free it */
	
	nick = irc_tolower(onick);

	if(this_DS_s==NULL) return NULL;
		
	if((np=DS_nick_lookup(this_DS_s,nick)) == NULL)
	{
		np = (struct node_nick_list *)malloc(sizeof(*np));
		if(np==NULL || (np->nick = nick)==NULL)
			return NULL;
		hashval = DS_nick_hash(nick);
		np->next = this_DS_s->DS_nick[hashval];
		this_DS_s->DS_nick[hashval] = np;
	}
	else
		free(np->onick);
	if((np->onick = strdup(onick)) == NULL)
		return NULL;
	np->host = NULL;
	np->realname = NULL;
	np->username = NULL;
	np->channell = NULL;
	return np;
}


void DS_nick_undef_destroy(struct node_nick_list *destroy)
{
	free(destroy->nick);
	free(destroy->onick);
	free(destroy);
}

bool DS_nick_undef(struct node_s_list *this_DS_s, char *onick)
{
	/* onick because we take care in this function of low/high issue */
	struct node_nick_list *np, *save;
	unsigned hashval;
	char *nick;

	nick = irc_tolower(onick);

	hashval = DS_nick_hash(nick);

	if(this_DS_s->DS_nick[hashval]==NULL) return false;

	if(strcmp(nick,this_DS_s->DS_nick[hashval]->nick)==0)
	{
		save=this_DS_s->DS_nick[hashval];
		this_DS_s->DS_nick[hashval]=save->next;
		DS_nick_undef_destroy(save);
		return true;
	}
	else
	{	
		np=this_DS_s->DS_nick[hashval];
		while(np->next != NULL)
		{
			if(strcmp(nick, np->next->nick) == 0)
			{
				save=np->next;
				np->next=save->next;
				DS_nick_undef_destroy(save);
				return true;
			}
			np = np->next;
		}
		return false;
	}
}

void DS_nick_giveall(struct node_s_list *this_DS_s)
{
	int i;
	struct node_nick_list *np;
	
	printf("%s\n","#### DS_nick_giveall ####");
	for(i=0;i<DS_NICK_HASHSIZE;i++)
	{
		np = this_DS_s->DS_nick[i];
		while(np!=NULL)
		{
		printf("%d - %s - %s\n",i,np->nick,np->onick);
		np=np->next;
		}
	}	
}

/* ######################################################## */
/* for the CHANNELS                                         */

struct node_channel_list *DS_channel_lookup(struct node_s_list *this_DS_s, char *s)
{
	struct node_channel_list *np;
	
	for(np=this_DS_s->DS_channel; np != NULL; np = np->next)
		if(strcmp(s, np ->name) == 0)
			return np; /* do issa ja */
	return NULL; /* verdammt */
}

struct node_channel_list *DS_channel_install(struct node_s_list *this_DS_s, char *name)
{
	struct node_channel_list *np;
	
	if((np=DS_channel_lookup(this_DS_s,name)) == NULL)
	{
		np = (struct node_channel_list *)malloc(sizeof(*np));
		if(np==NULL || (np->name = strdup(name))==NULL)
			return NULL;
		np->next = this_DS_s->DS_channel;
		np->nickl = NULL;
		np->topic = NULL;
		this_DS_s->DS_channel = np;
	}
/*	else
		free((void *) np->defn);
	if((np->defn = strdup(defn)) == NULL)
		return NULL;*/
	return np;
}

int DS_channel_undef(struct node_s_list *this_DS_s, char *name)
{
	struct node_channel_list *np, *save;

	if(strcmp(name,this_DS_s->DS_channel->name)==0)
	{
		save=this_DS_s->DS_channel;
		this_DS_s->DS_channel=save->next;
		free(save);
		return 1;
	}
	else
	{	
		np=this_DS_s->DS_channel;
		while(np->next != NULL)
		{
			if(strcmp(name, np->next->name) == 0)
			{
				save=np->next;
				np->next=save->next;
				free(save);
				return 1;
			}
			np = np->next;
		}
		return 0;
	}
}

void DS_channel_giveall(struct node_s_list *this_DS_s)
{
	struct node_channel_list *np;
	
	np = this_DS_s->DS_channel;

	printf("%s\n","#### DS_channel_giveall ####");

	while(np!=NULL)
	{
		printf("[%s] ",np->name);
		np=np->next;
	}
	printf("\n");
}

/* #####################################
   for the list of nicks in the channel data structure */

/* add "nick" to "channel", the nick has allready to exist in the NICK DS */
bool DS_channel_nick_add(struct node_s_list *this_DS_s, char *channel, char *nick)
{
	struct node_channel_list *putthere;
	struct node_nick_list *putthis;
	struct node_channel_nick_list *np;
	char *nick_tolower;

	if(this_DS_s==NULL) return false;
	if(channel==NULL) return false;
	if(nick==NULL) return false;

	nick_tolower = irc_tolower(nick);
	
	putthere = DS_channel_lookup(this_DS_s,channel);
	if(putthere==NULL) return false;

	putthis = DS_nick_lookup(this_DS_s,nick_tolower);
	if(putthis==NULL)
	{
		printf("DS_channel_nick_add: no such nick\n");
		return false;
	}
	
	np = (struct node_channel_nick_list *)malloc(sizeof(*np));
	if(np==NULL || (np->ptr = putthis)==NULL)
		return false;
	np->next = putthere->nickl;
	putthere->nickl = np;
	
	free(nick_tolower);	

	return true;
}

bool DS_channel_nick_del(struct node_s_list *this_DS_s, char *channel, char *nick)
{
	struct node_channel_list *delthere;
	struct node_channel_nick_list *np, *save;
	char *nick_tolower;

	if(this_DS_s==NULL) return false;
	if(channel==NULL) return false;
	if(nick==NULL) return false;

	nick_tolower = irc_tolower(nick);
	
	delthere = DS_channel_lookup(this_DS_s,channel);
	if(delthere==NULL) return false;

	if(strcmp(nick_tolower,delthere->nickl->ptr->nick)==0)
	{
		save=delthere->nickl;
		delthere->nickl=save->next;
		free(save);
		free(nick_tolower);
		return true;
	}
	else
	{	
		np=delthere->nickl;
		while(np->next != NULL)
		{
			if(strcmp(nick_tolower, np->next->ptr->nick) == 0)
			{
				save=np->next;
				np->next=save->next;
				free(save);
				free(nick_tolower);
				return true;
			}
			np = np->next;
		}
		free(nick_tolower);
		return false;
	}
}

/* this function is intended to give us a space separated list
with the nicks in the given channel or NULL if something goes wrong */
char *DS_channel_nick_list(struct node_s_list *this_DS_s, char *channel)
{
	struct node_channel_list *this_channel; /* pointer to the given channel */
	struct node_channel_nick_list *np; /* points to the actual working structure */
	char *s; /* is used to point to every byte copied to the output */
	char *output; /* this is the pointer to the "string" this function returns */
	int outpos=0; /* actual position in the output string */
	int outsize=1000; /* contain the size of *output */
	int incrementsize=1000; /* we increment in 1000 byte steps */
	int trash=100; /* if less memory is back for output, ne will be allocated */
		/* that does also mean, that we've got really a problem if a nick is bigger
		than 100 byte, i assume that this will NOT happen */

	/* if this_DS_s is NULL we've really got a problem */
	if(this_DS_s==NULL) return NULL;
	/* if channel is NULL that's a problem */
	if(channel==NULL) return NULL;

	/* no lets get a pointer to channel struct and look if it's there */
	this_channel = DS_channel_lookup(this_DS_s, channel);
	if(this_channel==NULL) return NULL;

	/* it's just shorter */
	np = this_channel->nickl;

	printf("%s\n",np->ptr->onick);

	/* if here np == NULL the list is empty */
	if(np == NULL) return NULL;

	/* we malloc a first portion of memory */
	if((output = (char *)malloc(outsize)) == NULL) return NULL;

	/* if np == NULL we've reached the end of the list */
	while(np != NULL)
	{
		/* we point s to the string containing the orginal nick
		(the case sensitive one) and write *s byte for byte to
		output while incrementing outpos, when s points to '\0'
		we have reached the end of the string */
		for(s=np->ptr->onick;*s!='\0';s++) output[outpos++] = *s;
		/* segregate with ' ' the last have to be removed at the end */
		output[outpos++] = ' ';

		/* check if enough unused bytes are allocated */
		if( (outsize-outpos) <= trash)
		{
			/* if not we increase the size of output with incrementsize */
			outsize = outsize + incrementsize;
			/* and realloc with this size */
			output = (char *)realloc(output,outsize);
		}
		/* jump to ne next structure in the list */
		np = np->next;
	}
	
	if(outpos>0)
	{
		/* ok we've 1 or more nicks, so the last char after the last nick
		is ' ', we overwrite it easily with '\0' */
		/* outpos points at the moment to the char after the ' '*/
		output[outpos-1]=0;
	}
	/* absolutly our output need outpos bytes, so lets realloc */	
	output = (char *)realloc(output,outpos);

	/* and return... */
	return output;
}



/*#################################################################*/


/* #####################################
   for the list of channels in the nick data structure */

/* add "channelchar" to "nickchar", the channel has allready to exist in the CHANNEL DS */
bool DS_nick_channel_add(struct node_s_list *this_DS_s, char *nickchar,char *channelchar)
{
	struct node_nick_list *putthere;
	struct node_channel_list *putthis;
	struct node_nick_channel_list *np;
	char *nick_tolower;

	if(this_DS_s==NULL) return false;
	if(nickchar==NULL) return false;
	if(channelchar==NULL) return false;

	nick_tolower = irc_tolower(nickchar);
	
	putthere = DS_nick_lookup(this_DS_s,nick_tolower);
	if(putthere==NULL) return false;

	putthis = DS_channel_lookup(this_DS_s,channelchar);
	if(putthis==NULL)
	{
		printf("DS_nick_channel_add: no such channel\n");
		return false;
	}
	
	np = (struct node_nick_channel_list *)malloc(sizeof(*np));
	if(np==NULL || (np->ptr = putthis)==NULL)
		return false;
	np->next = putthere->channell;
	putthere->channell = np;
	
	free(nick_tolower);	

	return true;
}

bool DS_nick_channel_del(struct node_s_list *this_DS_s, char *nickchar, char *channelchar)
{
	struct node_nick_list *delthere;
	struct node_nick_channel_list *np, *save;
	char *nick_tolower;

	if(this_DS_s==NULL) return false;
	if(nickchar==NULL) return false;
	if(channelchar==NULL) return false;

	nick_tolower = irc_tolower(nickchar);

	delthere = DS_nick_lookup(this_DS_s,nick_tolower);
	if(delthere==NULL) return false;

	free(nick_tolower);

	/*ok*/

	if(strcmp(channelchar,delthere->channell->ptr->name)==0)
	{
		save=delthere->channell;
		delthere->channell=save->next;
		free(save);
		return true;
	}
	else
	{	
		np=delthere->channell;
		while(np->next != NULL)
		{
			if(strcmp(channelchar, np->next->ptr->name) == 0)
			{
				save=np->next;
				np->next=save->next;
				free(save);
				return true;
			}
			np = np->next;
		}
		return false;
	}
}

/* this function is intended to give us a space separated list
with the nicks in the given channel or NULL if something goes wrong */
char *DS_nick_channel_list(struct node_s_list *this_DS_s, char *channel)
{
	struct node_channel_list *this_channel; /* pointer to the given channel */
	struct node_channel_nick_list *np; /* points to the actual working structure */
	char *s; /* is used to point to every byte copied to the output */
	char *output; /* this is the pointer to the "string" this function returns */
	int outpos=0; /* actual position in the output string */
	int outsize=1000; /* contain the size of *output */
	int incrementsize=1000; /* we increment in 1000 byte steps */
	int trash=100; /* if less memory is back for output, ne will be allocated */
		/* that does also mean, that we've got really a problem if a nick is bigger
		than 100 byte, i assume that this will NOT happen */

	/* if this_DS_s is NULL we've really got a problem */
	if(this_DS_s==NULL) return NULL;
	/* if channel is NULL that's a problem */
	if(channel==NULL) return NULL;

	/* no lets get a pointer to channel struct and look if it's there */
	this_channel = DS_channel_lookup(this_DS_s, channel);
	if(this_channel==NULL) return NULL;


	/* it's just shorter */
	np = this_channel->nickl;

	printf("%s\n",np->ptr->onick);

	/* if here np == NULL the list is empty */
	if(np == NULL) return NULL;

	/* we malloc a first portion of memory */
	if((output = (char *)malloc(outsize)) == NULL) return NULL;


	/* if np == NULL we've reached the end of the list */
	while(np != NULL)
	{
		/* we point s to the string containing the orginal nick
		(the case sensitive one) and write *s byte for byte to
		output while incrementing outpos, when s points to '\0'
		we have reached the end of the string */
		for(s=np->ptr->onick;*s!='\0';s++) output[outpos++] = *s;
		/* segregate with ' ' the last have to be removed at the end */
		output[outpos++] = ' ';

		/* check if enough unused bytes are allocated */
		if( (outsize-outpos) <= trash)
		{
			/* if not we increase the size of output with incrementsize */
			outsize = outsize + incrementsize;
			/* and realloc with this size */
			output = (char *)realloc(output,outsize);
		}
		/* jump to ne next structure in the list */
		np = np->next;
	}
	
	if(outpos>0)
	{
		/* ok we've 1 or more nicks, so the last char after the last nick
		is ' ', we overwrite it easily with '\0' */
		/* outpos points at the moment to the char after the ' '*/
		output[outpos-1]=0;
	}
	/* absolutly our output need outpos bytes, so lets realloc */	
	output = (char *)realloc(output,outpos);

	/* and return... */
	return output;
}



/*#################################################################*/





char *irc_tolower(char *input)
{
	char *s, *output;
	
	output = strdup(input);
	for(s=output; *s != '\0'; s++)
		if( (*s>=65) && (*s<=93) ) *s = *s + 32;
	return output;
}



#ifdef ALONE

int main(void)
{
	int i;
	DS_s = (struct node_s_list *)malloc(sizeof(*DS_s));
	DS_s->DS_channel=NULL;
	for(i=0;i<DS_NICK_HASHSIZE;i++) DS_s->DS_nick[i]=NULL;

	DS_channel_giveall(DS_s);
	DS_channel_install(DS_s, "#lala1");
	DS_channel_install(DS_s, "#lala2");
	DS_channel_install(DS_s, "#lala3");
	DS_channel_install(DS_s, "#lala4");
	DS_channel_install(DS_s, "#lala5");
	DS_channel_giveall(DS_s);
	DS_channel_undef(DS_s, "#lala3");
	DS_channel_giveall(DS_s);
	DS_channel_undef(DS_s, "#lala5");
	DS_channel_giveall(DS_s);
	DS_channel_undef(DS_s, "#lala1");
	DS_channel_giveall(DS_s);
	DS_channel_undef(DS_s, "#lala2");
	DS_channel_giveall(DS_s);
	DS_channel_undef(DS_s, "#lala4");
	DS_channel_giveall(DS_s);

	printf("%s\n","nick_install...");
	DS_nick_install(DS_s,"HaNs");
	printf("%s\n","HaNs installed");
	DS_nick_install(DS_s,"hAllo");
	DS_nick_install(DS_s,"flAucsh");
	DS_nick_install(DS_s,"Gecko");        /* delicty A */
	printf("%s\n","nick_giveall...");
	DS_nick_giveall(DS_s);
	printf("ok\n");
	DS_nick_undef(DS_s, "gecko");          /* delicty A */
	DS_nick_giveall(DS_s);
	printf("ok2\n");

	
	DS_channel_install(DS_s, "#ssc");
	if(!DS_channel_nick_add(DS_s, "#ssc","Gecko")) printf("can not add Gecko\n");
	DS_channel_nick_add(DS_s, "#ssc","hallo");
	printf("LIST: %s\n",DS_channel_nick_list(DS_s, "#ssc"));
	printf("ok3\n");
	DS_channel_nick_add(DS_s, "#ssc","hans");
	DS_channel_nick_del(DS_s, "#ssc","hallo");
	printf("LIST: %s\n",DS_channel_nick_list(DS_s, "#ssc"));
	DS_channel_nick_add(DS_s, "#ssc","hallo");
	printf("LIST: %s\n",DS_channel_nick_list(DS_s, "#ssc"));

/*	test=DS_channel_nick_list("#ssc");
	printf("%s ",test);
	printf("\n");
*/
	exit(0);
}
#endif