/* 
    SIT - SIS is tiny, intended to be a IRC Client for secure communication
	on top of the IRC protocol as described in RFC 1459
    Copyright (C) 2000  Christian Leber

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    e-mail: c_l@bigfoot.com
    I hope this is GPL compliant, if not please contact me.

    This is 'public domain' in the sense of the Wassenaar treaty.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <gmp.h>
#include <fcntl.h>
#include "rsa.h"
#include "big_numb.h"
#include "rsakeygenarg.h"



/*
void clrsa_probab_prime(mpz_t output,int bit_length)
{
	mpz_t pprime;
	
	mpz_init(pprime);
	big_pprime(pprime,bit_length,1);

	while(mpz_probab_prime_p(pprime,10)==0)
		mpz_add_ui(pprime,pprime,2);
		
	mpz_set(output,pprime);
	mpz_clear(pprime);
}
*/

int clrsa_create_key_pair(int bitlength, clrsa_public *rsa_public, clrsa_privat *rsa_privat)
{
    mpz_t rsa_p, rsa_q, rsa_n, rsa_n2, rsa_e, rsa_d, temp;

    mpz_init(rsa_p);
    mpz_init(rsa_q);
    mpz_init(rsa_n);
    mpz_init(rsa_n2);
    mpz_init(rsa_e);
    mpz_init(rsa_d);
    mpz_init(temp);

	/* first we need 2 big primes p and q */
	printf("%s\n","[1/9] - fetching the random prime, p");	
	clrsa_probab_prime(rsa_p,bitlength);
	printf("%s\n","[2/9] - fetching the random prime, q");	
	clrsa_probab_prime(rsa_q,bitlength);

	printf("%s\n","[3/9] - check that p!=q");	
	while(mpz_cmp(rsa_p,rsa_q)==0)
	{
  		clrsa_probab_prime(rsa_q,bitlength);
	}
	
	/* n  = p * q */
	printf("%s\n","[4/9] - calculate n = p * q");	
    mpz_mul(rsa_n,rsa_p,rsa_q);

	/* n2 = (p - 1) * (q - 1) */
	printf("%s\n","[5/9] - calculate n2 = (p-1) * (q-1)");	
    mpz_sub_ui(rsa_n2,rsa_p,1);
    mpz_sub_ui(temp,rsa_q,1);
    mpz_mul(rsa_n2,rsa_n2,temp);
    
	/* rsa_e has to be a random integer, but not neccessary a prime
	in the range from 1 < rsa_e < rsa_n2 */
	printf("%s\n","[6/9] - get random 64 bit integer e");	
	big_pprime(rsa_e,64,0);

	printf("%s\n","[7/9] - secure that the gcd from e and n2 is 1");	
	again:
	mpz_gcd (temp, rsa_e, rsa_n2);
	if (mpz_cmp_ui (temp, 1) != 0)
	{
		mpz_add_ui (rsa_e, rsa_e, 1);
		goto again;
	}

	printf("%s\n","[8/9] - calculate d (n and d are the privat key)");	
	
	mpz_gcdext (temp, rsa_d, 0, rsa_e, rsa_n2);
	mpz_tdiv_r (rsa_d, rsa_d, rsa_n2);
	if (mpz_sgn (rsa_d) < 0) mpz_add (rsa_d, rsa_d, rsa_n2);

	mpz_init_set(rsa_public->e,rsa_e);
	mpz_init_set(rsa_public->n,rsa_n);
	mpz_init_set(rsa_privat->d,rsa_d);
	mpz_init_set(rsa_privat->n,rsa_n);

	printf("%s\n","[9/9] - key calculation finished");	

    mpz_clear(rsa_p);
    mpz_clear(rsa_q);
    mpz_clear(rsa_n);
    mpz_clear(rsa_n2);
    mpz_clear(rsa_e);
    mpz_clear(rsa_d);
    mpz_clear(temp);
	
	return 0;
}

int clrsa_test(clrsa_public *rsa_public, clrsa_privat *rsa_privat)
{
	mpz_t test_source, test_encrypt, test_decrypt;
	unsigned char *test_raw;
	int i,length;

	mpz_init(test_source);
	mpz_init(test_encrypt);
	mpz_init(test_decrypt);

	length = mpz_sizeinbase(rsa_public->n,2);	
	length = length / 8;
	length = length - 1;
	test_raw = (unsigned char *)malloc(length);
	
	for(i=0;i<length;i++)
	{
		test_raw[i] = rand() % 256;
	}
	
    for (i=length-1;i>=0;i--)	{
		mpz_mul_2exp (test_source, test_source, 8);
		mpz_add_ui (test_source, test_source, test_raw[i]);
    }


	/* first encrypt */
	mpz_powm(test_encrypt,test_source,rsa_public->e,rsa_public->n);
	/* then decrypt */
	mpz_powm(test_decrypt,test_encrypt,rsa_privat->d,rsa_privat->n);

/*	printf("\n");
	printf("test_source: ");
	mpz_out_str(stdout, 10, test_source);
	printf("\n");
	printf("test_encrypt: ");
	mpz_out_str(stdout, 10, test_encrypt);
	printf("\n");
	printf("test_decrypt: ");
	mpz_out_str(stdout, 10, test_decrypt);
	printf("\n");
*/
	/* now lets compare if we've encryptet and decrypted right */
	if(mpz_cmp(test_source,test_decrypt)==0) return 1;
	return 0;
}

int main(int argc, char **argv)
{
	clrsa_public rsa_public;
	clrsa_privat rsa_privat;
	int length_of_pq, fp, string_length;
	char *fname_public, *fname_privat;
	FILE *fstream;
	mpz_t temp1,temp2;
	int sstringsize_max=4096; /* a string can't be longer than this, we've to check */
	char st1[4097],st2[4097];
	Cmdline *cmd; /* Cmdline is a structure for the possible arguments */ 

	/* let the command line parser fill up his structure */
	cmd = parseCmdline(argc, argv); 

	/* If there is the -h argument give the usage message and exit */
	if(cmd->helpP!=0){usage();}

	if(cmd->lengthP!=0)
	{
		length_of_pq=cmd->length;
	}
	else
	{
		length_of_pq=512;
	}

	if(length_of_pq<=40) /* FIXME  this must be 512 or so */
	{
		printf("%d %s\n",length_of_pq,"bit length... this is nice...");
		printf("%s\n","try again!");
		exit(1);
	}
	else if(length_of_pq<512)
	{
		printf("%s\n","this is not a big key, it is a bit to few!");
	}	
	else if(length_of_pq<1536)
	{
		printf("%s\n","this key length is ok!");
	}
	else
	{
		printf("%s\n","this key is getting really big, it will take a long time... it's your time/live... !");
	}

		
	if(cmd->bothP!=0)
	{
		string_length = strlen(cmd->both);
		fname_privat = (char *)malloc(string_length+13);
		fname_public = (char *)malloc(string_length+13);
		strcpy(fname_privat,cmd->both);
		strcat(fname_privat,".privat.rsa");
		strcpy(fname_public,cmd->both);
		strcat(fname_public,".public.rsa");		
	}
	else
	{
		fname_privat = (char *)malloc(14);
		fname_public = (char *)malloc(14);
		strcpy(fname_privat,"my.privat.rsa");
		strcpy(fname_public,"my.public.rsa");
	}	

	if(cmd->publicnameP!=0)
	{
		string_length = strlen(cmd->publicname);
		fname_public = (char *)realloc(fname_public, string_length+1);
		strcpy(fname_public,cmd->publicname);
	}	

	if(cmd->privatnameP!=0)
	{
		string_length = strlen(cmd->privatname);
		fname_public = (char *)realloc(fname_privat, string_length+1);
		strcpy(fname_privat,cmd->privatname);
	}	

	if( (fp=open(fname_privat,O_RDONLY)) !=-1)
	{
		printf("%s%s%s\n","file (",fname_privat,") is allready there, delete it or take another name");
		exit(1);
	}
	if( (fp=open(fname_public,O_RDONLY)) !=-1)
	{
		printf("%s%s%s\n","file (",fname_public,") is allready there, delete it or take another name");
		exit(1);
	}
	

	clrsa_create_key_pair(length_of_pq,&rsa_public,&rsa_privat);

	printf("%s","testing encryption... ");
	fflush(stdout);
	if(clrsa_test(&rsa_public,&rsa_privat)==1) /* if the return 1, it's allright */
	{
		printf("%s\n","OK!");
	}
	else
	{
		printf("%s\n","failed !!!!");
		printf("\n%s\n","######################## ERROR ############################");
		printf("%s\n","there must be a really serious problem, please read the FAQ");
	}

	mpz_init(temp1);
	mpz_init(temp2);

	if(mpz_sizeinbase(rsa_public.n,10) >= sstringsize_max)
	{
		printf("%s %d %s","up to this all is ok, but n is to long, the maximum are:",sstringsize_max,"decimals");
		exit(1);
	}
	

	printf("%s %s %s","Writing PUBLICKEY to",fname_public,"... ");
	fflush(stdout);	
	fstream=fopen(fname_public,"w");
	fprintf(fstream,"%s\n","PUBLICKEY_please_do_not_change_anything_in_this_file_by_hand");
	mpz_out_str(fstream,10,rsa_public.n);
	fprintf(fstream,"\n");
	mpz_out_str(fstream,10,rsa_public.e);
	fclose(fstream);

	fstream=fopen(fname_public,"r");
	if(fscanf(fstream,"%*s\n%s\n%s",st1,st2)!=2) printf("%s\n","#error#");
	mpz_set_str(temp1,st1,10);
	mpz_set_str(temp2,st2,10);
	if(mpz_cmp(rsa_public.n,temp1)!=0 || mpz_cmp(rsa_public.e,temp2)!=0)
	{
		printf("%s\n","failed!");
		exit(1);
	}
	else printf("%s\n","OK!");
	
	fclose(fstream);

	printf("%s %s %s","Writing PRIVATKEY to",fname_privat,"... ");
	fflush(stdout);	
	fstream=fopen(fname_privat,"w");
	fprintf(fstream,"%s\n","PRIVATKEY_please_do_not_change_anything_in_this_file_by_hand");
	mpz_out_str(fstream,10,rsa_privat.n);
	fprintf(fstream,"\n");
	mpz_out_str(fstream,10,rsa_privat.d);
	fclose(fstream);

	fstream=fopen(fname_privat,"r");
	if(fscanf(fstream,"%*s\n%s\n%s",st1,st2)!=2) printf("%s\n","#error#");
	mpz_set_str(temp1,st1,10);
	mpz_set_str(temp2,st2,10);
	if(mpz_cmp(rsa_privat.n,temp1)!=0 || mpz_cmp(rsa_privat.d,temp2)!=0)
	{
		printf("%s\n","failed!");
		exit(1);
	}
	else printf("%s\n","OK!");
	
	fclose(fstream);

	exit(0);
}
