/* 
    SIT - SIS is tiny, intended to be a IRC Client for secure communication
	on top of the IRC protocol as described in RFC 1459
    Copyright (C) 2000  Christian Leber

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    e-mail: c_l@bigfoot.com
    I hope this is GPL compliant, if not please contact me.
*/

#define SIT_VERSION "v.013 dev 16"
#define DEBUG

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>

#ifndef __EMX__
   #include <arpa/inet.h>
#endif
#ifdef __EMX__
	#include <sys/select.h>
 	#include <termios.h>
 	#define IDEFAULT    0x8000
#endif
   
#include <string.h>
#include <netdb.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <signal.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <stdarg.h>
#include <fcntl.h>
#include <ctype.h>
#include "../blowfish/common.h"
#include "../blowfish/blowfish.h"
#include "cmdarg.h"
#include "databox.h"
#include "../configsit.h"
#include "dstruct.h"
#include "out.h"
#include "all.h"

#if defined __EMX__
	#define DEFAULT_OS "OS/2, EMX"
 	struct termios tio;
#elif defined __CYGWIN__
	#define DEFAULT_OS "GNU/Cygwin32"
#else
	#define DEFAULT_OS "GNU/Linux"
#endif

#define max(a,b)	((a) > (b) ? (a) : (b))

#define CONFIG_SIT_SOCKBUF_SIZE 1024

WORD08 BLOWFISH_KEY[128];
WORD32 BLOWFISH_KEY_SCHED[1042];
int BLOWFISH_INIT=0;

static struct DS_blowfish *BLOWFISH_TEMP;
static int out_protocol=1;

static struct node_s_list *DS_s; /* this holds all information about the
	server we connect to and data structures witch belong to this */

/* now is all in the correct order */
char *getnick(char *input_from);

bool fserver_in(char *input);
bool fserver_in2(char *input);
bool fserver_in_notice(char *input_from, char *input_dest, char *input_msg);
bool fserver_in_privmsg(char *input_from, char *input_dest, char *input_msg);
bool fserver_in_privmsg_decrypt(char *input_from_nick, char *input_dest, char *input_msg);
bool fserver_in_minimal(char *input);

bool nicks_from_353(char *argu_mess, char *channel);

void SIT_blowfish_init(void);
char *SIT_blowfish_encrypt(char *input);
char *SIT_blowfish_decrypt(char *input);
bool SIT_Connect(const char *,const int);

char *parser_dp_cut_left(const char *input); 
char *parser_spacecut_left(const char *input);
char *parser_autocut_right(char *input);
char *parser_cut_left(char *input, int number);

char *unprint(char *input);

void SIT_parse_input(char *input);
bool SIT_parse_input_cmd(char *input);

void SIT_select(void);
void SC_help(const char *cmd);

void socket_to_buffer(void);
bool buffer_to_msg(void);
bool SIT_tcpSend(char *input);
void SIT_printf_to_tcpSend(char *format, ...);

char *getnick(char *input_from)
{
	int maxnicklength=32;
	char *s,*output;
	int opos=0;
	
	output = (char *)malloc(maxnicklength+1);

	s = input_from;
	while( (*s!='\0') && (*s!='!') && (opos < maxnicklength-1) )
	{
		output[opos++] = *s;
		s++;
	}
	output[opos]=0;

/*	if(sscanf(input_from,"%30[^!]!%*[-+=~]%*[^@]@%*s",output) != 1)
		if(sscanf(input_from,"%30[^!]!^%*[^@]@%*s",output) != 1)
			if(sscanf(input_from,"%30[^!]!%*[^@]@%*s",output) != 3)
				return NULL;*/
	return output;
}

bool fserver_in(char *input)
{
#ifdef DEBUG
/*	FILE *fp;*/
#endif
	char *msg_from;
	char *msg_cmd;
	char *msg_dest;
	char *msg_mess;

	bool hit=false;

	msg_from = (char *)malloc(513);
	msg_cmd  = (char *)malloc(11);
	msg_dest = (char *)malloc(513);
	msg_mess = (char *)malloc(513);

	if(sscanf(input,":%512s %10s %512s :%512[^\n]",msg_from,msg_cmd,msg_dest,msg_mess) != 4)
		{return false;}

	if(strcmp(msg_cmd,"PRIVMSG")==0)
	{
		if(fserver_in_privmsg(msg_from,msg_dest,msg_mess))
			hit=true;
	} 

	else if(strcmp(msg_cmd,"NOTICE")==0)
	{
		if(fserver_in_notice(msg_from,msg_dest,msg_mess))
			hit=true;
	} 

	else if( (strcmp(msg_cmd,"001")==0 || strcmp(msg_cmd,"002")==0 ||
			strcmp(msg_cmd,"003")==0 || strcmp(msg_cmd,"251")==0 ||
			strcmp(msg_cmd,"255")==0 || strcmp(msg_cmd,"372")==0 ||
			strcmp(msg_cmd,"375")==0) && strcmp(msg_dest,DS_s->NICKNAME)==0 )
	{
		out_server(out_protocol,0,unprint(msg_mess));
		hit=true;
	} 

	/* The following Stuff is for the EUIRC, it's NOT in described in
		the RFC, and there is NO paper so I'm not sure if I interpret
		the Stuff right */
	else if( (strcmp(msg_cmd,"265")==0 || strcmp(msg_cmd,"266")==0)
			 && strcmp(msg_dest,DS_s->NICKNAME)==0 )
	{
		out_server(out_protocol,0,unprint(msg_mess));
		hit=true;
	} 


	else if(strcmp(msg_cmd,"376")==0)
	{
		out_server(out_protocol,0,unprint(msg_mess));
		SIT_printf_to_tcpSend("JOIN %s\r\n",DS_s->PCHANNEL);
		hit=true;
	} 

	if(hit){return true;}
	else {return false;}
}

bool fserver_in2(char *input)
{
#ifdef DEBUG
/*	FILE *fp;*/
#endif
	char *msg_from;
	char *msg_cmd;
	char *msg_dest;
	char *msg_add; /* information like the number of channels */
	char *msg_mess;
	char *buff_1;
	bool hit=false;

	msg_from = (char *)malloc(513);
	msg_cmd  = (char *)malloc(11);
	msg_dest = (char *)malloc(513);
	msg_add  = (char *)malloc(513);
	msg_mess = (char *)malloc(513);
	buff_1 = (char *)malloc(1025);

	if(sscanf(input,":%512s PART %512s :%512[^\n]",msg_from,msg_dest,msg_mess) == 3)
	{
		snprintf(buff_1,1024,"<<< %s leaves %s (%s)",msg_from,msg_dest,msg_mess);
		out_raw(out_protocol,unprint(buff_1));
		/*FIXME!*/
		hit=true;
	}

	else if(sscanf(input,":%512s MODE %512s %512[^\n]",msg_from,msg_dest,msg_mess) == 3)
	{
		snprintf(buff_1,1024,"--[%s]-MODE-[%s]-: %s",msg_dest,msg_from,msg_mess);
		out_raw(out_protocol,unprint(buff_1));
		hit=true;
	}

	else if(sscanf(input,":%512s NICK :%512[^\n]",msg_from,msg_dest) == 2)
	{
		snprintf(buff_1,1024,"*** %s changed nick to: %s",msg_from,msg_dest);
		out_raw(out_protocol,unprint(buff_1));
		/*FIXME!*/
		hit=true;
	}

	else if(sscanf(input,":%512s TOPIC %512s :%512[^\n]",msg_from,msg_dest,msg_mess) == 3)
	{
		snprintf(buff_1,1024,"--[%s]-TOPIC-[%s]: %s",msg_dest,msg_from,msg_mess);
		out_raw(out_protocol,unprint(buff_1));
		hit=true;
	}

	else if(sscanf(input,":%512s QUIT :%512[^\n]",msg_from,msg_mess) == 2)
	{
		snprintf(buff_1,1024,"-<<<-- %s has quit IRC (%s)",msg_from,msg_mess);
		out_raw(out_protocol,unprint(buff_1));
		/*FIXME!*/
		hit=true;
	}

	else if(sscanf(input,":%512s KICK %512s %512s :%512[^\n]",msg_from,msg_add,msg_dest,msg_mess) == 4)
	{
		printf("!! %s\n",msg_dest);
		if(strcmp(msg_dest,DS_s->NICKNAME)==0)
		{
			snprintf(buff_1,1024,"-{}-{}-[ ...HU!!! WE were kicked... from %s by %s][%s]-{}-{}-",msg_add,msg_from,msg_mess);
			out_raw(out_protocol,unprint(buff_1));
			/*FIXME!*/
			hit=true;
		}
		else
		{
			snprintf(buff_1,1024,"--[ %s was kicked from %s by %s][%s]",msg_dest,msg_add,msg_from,msg_mess);
			out_raw(out_protocol,unprint(buff_1));
			/*FIXME!*/
			hit=true;
		}
	}

	else if(sscanf(input,":%512s 332 %512s %512s :%512[^\n]",msg_from,msg_dest,msg_add,msg_mess) == 4)
	{
		if(strcmp(msg_dest,DS_s->NICKNAME)==0)
		{
			snprintf(buff_1,1024,"--[%s]-TOPIC: %s",msg_add,msg_mess);
			out_raw(out_protocol,unprint(buff_1));
			hit=true;
		}
	}

	else if(sscanf(input,":%512s 353 %512s = %512s :%512[^\n]",msg_from,msg_dest,msg_add,msg_mess) == 4)
	{
		if(strcmp(msg_dest,DS_s->NICKNAME)==0)
		{
			snprintf(buff_1,1024,"--[%s]: %s",msg_add,msg_mess);
			out_raw(out_protocol,unprint(buff_1));
			nicks_from_353(msg_mess,msg_add);
		
			hit=true;
		}
	}

	else if(sscanf(input,":%512s 004 %512s %512[^\n]",msg_from,msg_dest,msg_mess) == 3)
	{
		if(strcmp(msg_dest,DS_s->NICKNAME)==0)
		{
			out_server(out_protocol,0,unprint(msg_mess));
			hit=true;
		}
	}

	else if(sscanf(input,":%512s JOIN %512s",msg_from,msg_dest) == 2)
	{
		if(msg_dest[0]==':') msg_dest = msg_dest + 1;
		/* check if WE join this channel */
		/* msg_from_nick = getnick(msg_from);*/
		if(strcmp(DS_s->NICKNAME,getnick(msg_from)) == 0)
		{
			DS_channel_install(DS_s,msg_dest);
		}
		snprintf(buff_1,1024,">>> %s joinss %s",msg_from,msg_dest);
		out_raw(out_protocol,unprint(buff_1));
		
		/*FIXME! jmd. hat den channel gejoint und muss eingetragen werden */
		DS_nick_install(DS_s,msg_from);
		
		hit=true;
	}

	else if(sscanf(input,":%512s %10s %512s %30s :%512[^\n]",msg_from,msg_cmd,msg_dest,msg_add,msg_mess) == 5)
	{
		if( (strcmp(msg_cmd,"252")==0 || strcmp(msg_cmd,"253")==0 ||
			strcmp(msg_cmd,"254")==0) && strcmp(msg_dest,DS_s->NICKNAME)==0 )
		{
			snprintf(buff_1,1024,"%s: %s",msg_mess,msg_add);
			out_server(out_protocol,0,unprint(buff_1));
			hit=true;
		}
		if(strcmp(msg_cmd,"366")==0 && strcmp(msg_dest,DS_s->NICKNAME)==0)
		{
			snprintf(buff_1,1024,"--[%s]: %s",msg_add,msg_mess);
			out_raw(out_protocol,unprint(buff_1));
			hit=true;
		}
	}

	/* non RFC shitt */	
	else if(sscanf(input,":%512s 005 %512s %512[^\n]",msg_from,msg_dest,msg_mess) == 3)
	{
		if(strcmp(msg_dest,DS_s->NICKNAME)==0)
		{
			out_server(out_protocol,0,unprint(msg_mess));
			hit=true;
		}
	}

	if(hit){return true;}
	else {return false;}
}


bool fserver_in_notice(char *input_from, char *input_dest, char *input_msg)
{
	char *from_nick; /* only the nick */
	char *from_user; /* only the username */
	char *from_host; /* only the host */
	int length;
	bool hit=false; /* at this point... :) */
	unsigned long pingtime;
	
	/* first the length, it is in every case enough */
	length = strlen(input_from)+1;
	from_nick = (char *)malloc(length);
	from_user = (char *)malloc(length);
	from_host = (char *)malloc(length);

	if(sscanf(input_from,"%512[^!]!%*[-+=~]%512[^@]@%s",from_nick,from_user,from_host) != 3)
		if(sscanf(input_from,"%[^!]!^%[^@]@%s",from_nick,from_user,from_host) != 3)
			if(sscanf(input_from,"%[^!]!%[^@]@%s",from_nick,from_user,from_host) != 3)
			{
				if(sscanf(input_from,"%s",from_nick) == 1)
				{
					from_user=NULL;
					from_host=NULL;
				}
				else
				{
					return false;
				}
			}
	
	if(strcmp(input_dest,DS_s->NICKNAME)==0)
	{
		if(sscanf(input_msg,"\001PING %ld\001",&pingtime)==1)
		{
		/* should be ok, but this could be a possibility of an attack
			if the number is fancy*/
			printf("[%s PING reply]: %ld secs\n",from_nick, time(NULL)-pingtime);
			hit=true;
		}
		else
		{	
			out_notice(out_protocol,1,from_nick,input_msg);
			hit=true;
		}
	}
	else if(strcmp(input_dest,"AUTH")==0)
	{
		out_notice(out_protocol,1,from_nick,input_msg);
		hit=true;
	}

	if(hit) return true;
	return false;
}


bool fserver_in_privmsg(char *input_from, char *input_dest, char *input_msg)
{
	char *from_nick; /* only the nick */
	char *from_user; /* only the username */
	char *from_host; /* only the host */
	int length;
	bool hit=false; /* at this point... :) */
	
	/* first the length, it is in every case enough */
	length = strlen(input_from)+1;
	from_nick = (char *)malloc(length);
	from_user = (char *)malloc(length);
	from_host = (char *)malloc(length);

	if(sscanf(input_from,"%[^!]!%*[-+=~]%[^@]@%s",from_nick,from_user,from_host) != 3)
		if(sscanf(input_from,"%[^!]!^%[^@]@%s",from_nick,from_user,from_host) != 3)
			if(sscanf(input_from,"%[^!]!%[^@]@%s",from_nick,from_user,from_host) != 3)
				return false;

/*FIXME!*/
/*FIXME!*/
	
	if(strcmp(input_dest,DS_s->NICKNAME)==0)
	{
		if( (strcmp(input_msg,"\001VERSION\001")==0)||
			(strcmp(input_msg,"\001version\001")==0) )
		{
			SIT_printf_to_tcpSend("NOTICE %s :\001VERSION SIT by ijuz # %s # %s\001\r\n",from_nick,
			SIT_VERSION,DEFAULT_OS);
			printf("%s %s\n","CTCP-VERSION-REQUEST from:", from_nick);
			hit=true;
		}
		else if(strncmp(input_msg,"\001PING",5)==0)
		{
			SIT_printf_to_tcpSend("NOTICE %s :%s\r\n",from_nick,input_msg);
			printf("%s %s\n","CTCP-PING from:", from_nick);
			hit=true;
		}
		else
		{
			if(input_msg[0]=='@')
			{
				fserver_in_privmsg_decrypt(from_nick, NULL, input_msg);
				hit=true;
			}
			else
			{
				out_pm(out_protocol,2,from_nick,NULL,input_msg);
				hit=true;
			}
		}
	}

	else
	{	
		if(input_msg[0]=='@')
		{
			fserver_in_privmsg_decrypt(from_nick, input_dest, input_msg);
			hit=true;
		}
		else
		{
			out_pm(out_protocol,1,from_nick,input_dest,input_msg);
			hit=true;
		}
	}

	if(hit) return true;
	return false;
}

/* this will take the PRIVMSG witch is either something like a query
	(then input_dest have to be NULL or	public in a channel, than input_dest
	have to be the channel name */
bool fserver_in_privmsg_decrypt(char *input_from_nick, char *input_dest, char *input_msg)
{
	char *msg;
	char *msg_dec;
	char *msg_out;
	
	if(strlen(input_msg)<5) return false;
	msg = input_msg + 1;

	msg_dec = SIT_blowfish_decrypt(msg);
	msg_out = (char *)malloc(1025);

	if(msg_dec!=NULL)
	{
		snprintf(msg_out,1024,"%s%s","@",unprint(msg_dec));
		if(input_dest==NULL) out_pm(out_protocol,2,input_from_nick,NULL,msg_out);
		else out_pm(out_protocol,1,input_from_nick,input_dest,msg_out);
	}
	else
	{
		snprintf(msg_out,1024,"!!! %s has sended an invalid kugelfischly message(%s)",input_from_nick,input_msg);
		if(input_dest==NULL) out_pm(out_protocol,2,input_from_nick,NULL,msg_out);
		else out_pm(out_protocol,1,input_from_nick,input_dest,msg_out);
	}
	return true;	
}


bool fserver_in_minimal(char *input)
{
	char *buf1;

	buf1 = (char *)malloc(strlen(input) + 1);
	
	if(input[0]!=':')
	{
		if(strncmp(input,"NOTICE ",7)==0)
		{
			buf1=parser_dp_cut_left(input);
			out_server(out_protocol,0,unprint(buf1));
			return true;
		}

		if(strncmp(input,"PING ",5) == 0)
		{
			buf1=parser_cut_left(input,6);
			SIT_printf_to_tcpSend("PONG %s\r\n",buf1);
			return true;
		}
		
		else
		{
			return false;
		}
	}
	return false;
}


/* this should add the nicks in the 353 message to the data structure from the channel
and create data structures for the nicks if they are not allready avaible */
bool nicks_from_353(char *argu_mess, char *channel)
{
	bool stillsearch=true;	
	char *nick;
	int nickpos;
	char *s;
	char *argu_mess2;

	nick = (char *)malloc(40);
	argu_mess2=parser_autocut_right(argu_mess);
	s=argu_mess2;
		
	while(stillsearch)
	{
		nickpos=0;
		for(;(*s!=0) && (*s!=32);s++)
		{
			nick[nickpos++]=*s;
		}
		if(*s==32) s++;
		else if(*s==0) stillsearch=false;
		nick[nickpos]=0;
		DS_nick_install(DS_s,nick);
	}
	return true;
}


void SIT_blowfish_init(void)
{
	int fp,i;
		
	if( (fp=open("blowfish.key",O_RDONLY)) == -1)
	{
		printf("can't open blowfish.key\n");
		printf("!!! Default KEY is used !!!\n");
		BLOWFISH_TEMP->blowfish_key=strdup("Dies ist ein zum testen, nicht sicher!");
		BLOWFISH_TEMP->blowfish_length=strlen(BLOWFISH_TEMP->blowfish_key);
	}
	else
	{
		i=read(fp,BLOWFISH_TEMP->blowfish_key,72);
		BLOWFISH_TEMP->blowfish_length=i;
		close(fp);
	}
	
	DS_blowfish_schedule(BLOWFISH_TEMP);
}
 
char *SIT_blowfish_encrypt(char *input)
{
	unsigned char daten[8];
	int i,j,ipos,length_input,length,length8;
	unsigned char *buff_1;
	char *output;
	DataBox zwisch;

	/* look if the blowfish schedule is allready created */
	if(!BLOWFISH_TEMP->blowfish_init){SIT_blowfish_init();}
	length_input = strlen(input);
	if(length_input==0){printf("shitt happens in blowfish_encrypt");}
	/* length8 is the number of 8 byte long parts, for blowfish */
	length8 = (length_input / 8);
	if( (length8 * 8) - length_input != 0) {length8 = length8 + 1;}
	/* length is the number of bytes toencrypt total */
	length = length8 * 8;

	/*allociate the needed amount of memory for buff1_*/

	buff_1 = (unsigned char *)malloc(length + 1);

	/*buff_1 is the "internal" version of input, now we transfer input and fill up
	the "empty" bytes with 32, this is dangerous, because a posible securoty issue
	this have to be changed */
	
	for(i=0;i<length_input;i++)	{buff_1[i] = input[i];}
	for(i=length_input;i<length;i++) {buff_1[i] = 32;} 

	/* working through every 8 byte part in buff_1 */
	for(j=0;j<length8;j++)
	{
		ipos = j * 8;
		for(i=0;i<8;i++)
		{
			daten[i]=buff_1[ipos+i];
		}
		blowfish_encrypt(daten,BLOWFISH_TEMP->blowfish_sched);
		for(i=0;i<8;i++)
		{
			buff_1[ipos+i]=daten[i];
		}
	}
	/* buff_1 contains now the encrypted data */

	zwisch.data = (unsigned char*)malloc(length);
	for(i=0;i<length;i++){zwisch.data[i] = buff_1[i];}
	zwisch.length = length;
	
	output = DataBox_to_base64(zwisch);
	
	free(buff_1);
	return output;
}


char *SIT_blowfish_decrypt(char *input)
{
	unsigned char daten[8];
	int i,j,ipos,length4,length8,length_input;
	char *output;
	DataBox zwisch;
	
	if(!BLOWFISH_TEMP->blowfish_init){SIT_blowfish_init();}	
	length_input = strlen(input);

	if(length_input == 0){return NULL;}
	length4 = length_input / 4;
	if(length4 * 4 != length_input) {return NULL;}

	zwisch = base64_to_DataBox(input);
	if(zwisch.data==NULL) return NULL;

	output = (char *)malloc(length4*3+1);

	for(i=0;i<length4*3;i++)output[i]=zwisch.data[i];

	length8 = (length4 * 3) / 8;

	for(j=0;j<length8;j++)
	{
		ipos = j * 8;
		for(i=0;i<8;i++)
		{
			daten[i]=output[ipos+i];
		}
		blowfish_decrypt(daten,BLOWFISH_TEMP->blowfish_sched);
		for(i=0;i<8;i++)
		{
			output[ipos+i]=daten[i];
		}
	}
		
	output[length4*3]=0;
	
	return output;
}


bool SIT_Connect(const char *Server,const int Port)
{
	struct sockaddr_in servaddr;
	struct hostent *HostAddr;
	
	if(DS_s->SOCKET != 0) close(DS_s->SOCKET);
	DS_s->SOCKET = 0;

	DS_s->SOCKBUF_POS = 0;
	DS_s->SOCKBUF[0] = 0;
  
	DS_s->SOCKET = socket(AF_INET, SOCK_STREAM, getprotobyname("tcp")->p_proto);
	if(DS_s->SOCKET == -1) return false;
	printf("DEBUG: %s\n","SIT_Connect: socket created");

	memset(&servaddr, 0, sizeof(servaddr));

	servaddr.sin_family = AF_INET;
	servaddr.sin_port = htons(Port);
	
	if(inet_addr(Server) != INADDR_NONE) 
	{
		servaddr.sin_addr.s_addr = inet_addr(Server);
	}
	else
	{
		HostAddr = (struct hostent *)gethostbyname(Server);
		if(HostAddr == NULL)
		{
			printf("DEBUG: %s\n","SIT_Connect; gethostbyname() failed");
			return false;
		}
		memcpy((void *)&servaddr.sin_addr, HostAddr->h_addr, HostAddr->h_length);
	}
	if(connect(DS_s->SOCKET, (struct sockaddr *)&servaddr, sizeof(servaddr)) == -1)
		return false;
	printf("DEBUG: %s\n","SIT_Connect: connected");
	return true;
}

char *parser_dp_cut_left(const char *input)
{
	int i = 0, pos = 0;
	char *output;
 	
	while(input[i] != ':') {i++;}
	output = (char *)malloc(strlen(input)-i+1);
	for(;i+1<=strlen(input);i++) {output[pos++] = input[i];}

 	return output;
}

char *parser_spacecut_left(const char *input)
{
	int i = 0, pos = 0;
	char *output;
 	
	while(input[i] == ' ') {i++;}
	output = (char *)malloc(strlen(input)+1-i);
	for(;i<=strlen(input);i++) {output[pos++] = input[i];}

 	return output;
}

char *parser_autocut_right(char *input)
{
	int i, max;
	char *output;
	
	max = strlen(input);
	while(((input[max]==0)||(input[max]==10)||(input[max]==32)) && max >= 0)
		{max=max-1;}
	output = (char *)malloc(max+2);
	if(max==-1) output[0]=0;
	else
	{
		for(i=0;i<=max;i++) {output[i] = input[i];}
		output[max+1]=0;
	}
	return output;
}

char *parser_cut_left(char *input,int number)
{
	int len, pos = 0, i;
	char *output;

	len = strlen(input);
	output = (char *)malloc(len+1);
	if(len<=number){output[0]=0; return output;}
	for(i=number;i<=len;i++){output[pos++] = input[i];}
	return output;
}


char *unprint(char *input)
{
	char *output;
	int pos=0, len, i;

	len=strlen(input);
	output = (char *)malloc(len+1);
	for (i = 0; i < len; i++)
	{
		if(input[i]>=32) {output[pos++]=input[i];}
	}
	output[pos]=0;
	return output;
}


void SIT_parse_input(char *input)
{
#ifdef DEBUG
	FILE *fp;
	fp = fopen("log.input","a");
	fprintf(fp,"\"%s\"\n",input);
	fclose(fp);
#endif

	/* remove all \n and #32 from the right side */
	input = parser_autocut_right(input);
	
	/* have a look if it's now a empty string */
	if(input[0]==0) {return;}
	
	if(input[0]=='@')
	{
		SIT_printf_to_tcpSend("PRIVMSG %s :@%s\r\n",DS_s->PCHANNEL,SIT_blowfish_encrypt(input+1));
	}
	else if(input[0]=='/')  /* sign for commands */
	{
		if(!SIT_parse_input_cmd(input+1))
			{printf("%s\n","not a valid command");}
	}
	else
	{
		SIT_printf_to_tcpSend("PRIVMSG %s :%s\r\n",DS_s->PCHANNEL,input);
	}
}


bool SIT_parse_input_cmd(char *input)
{
	char *cmd1; /* for the first part of the command */
	char *input1; /* all after the *cmd1 and another #32 */
	char *input1l; /* all after the *cmd1 and all #32 */
	char *cmd2; /*for the possible next part of the command
				if there is no second part it will be NULL*/
	char *input2; /* all after *cmd2 and another #32*/
	size_t leng=0; /* this is used to find #32 and #0...*/
	int problem=0; /* if it's bigger then 0 at the end....
					there was a problem */
	int i,imax; /* just for a for construct */

	/* now lets search for #32 or the end of our *input */
	while( input[leng]!=0 && input[leng]!=32 ){leng+=1;}
	if(leng==0){return 0;} /*then the first character is " " or \0*/

	/* we need some memory for our first part of the command */
	cmd1 = (char *)malloc(leng+1);
	
	/* now lets copy leng-bytes from *input to *cmd1, that is the
		first part of our command */
	memcpy(cmd1,input,leng);
	/* now we should terminate the string :)*/
	cmd1[leng]=0;

	/* the following sets alls characters in *cmd1 to a lower
		character, thanks to Toppas for this idea */
	imax=strlen(cmd1);
	for(i=0;i<imax;i++){cmd1[i]=tolower(cmd1[i]);}
	
	/* the next thing is pointer fun... i point it to the byte after the
		#32 we have found above... or to NULL if we've found the end of the string */

	if(input[leng]==0) {input1=NULL;}
	else
	{
		/*... if the next byte is 0... then there is nothing back :) */
		if(input[leng+1]==0){input1=NULL;}
			else{input1 = input + leng + 1;}
	}
	
	/* In the following and before I don't care about someting like "... 32 32 0"
		because parser_autocut_right have done it's job in SIT_parse_input
		We set a new pointer input1l to the first byte after some possibly
		existent 32's... this is neccessary so that "/msg foo bar" is the same
		like "msg     foo bar"		
	*/

	if(input1==NULL){input1l=NULL;}
	else
	{
		input1l=input1;
		while(input1l[0]==32){input1l+=1;}
	}

	/* now we search cmd2 and set a pointer to the rest, the rest
	is like above */
	if(input1l!=NULL)
	{
		leng=0;
		while(input1l[leng]!=0 && input1[leng]!=32 ){leng+=1;}
		cmd2 = (char *)malloc(leng+1);
		memcpy(cmd2,input1l,leng);
		cmd2[leng]=0;

		if(input1l[leng]==0) {input2=NULL;}
		else
		{
			if(input1l[leng+1]==0){input2=NULL;}
				else{input2 = input1l + leng + 1;}
		}
	}
	else
	{
		cmd2 = NULL;
		input2 =  NULL;
	}
		
	/* this is the great exit!!! ;-)) */
	if ( (strcmp(cmd1,"exit")==0) || (strcmp(cmd1,"quit")==0) )
	{
		/*
			>> man printf
			>??? --verbose!!!!
			go die!!!		
		*/
		printf("---\nSIT: going down\n");

		if(input1!=NULL)
		{
			SIT_printf_to_tcpSend("QUIT :%s\r\n",input1);
		}
		else
		{		
			SIT_printf_to_tcpSend("QUIT :%s (%s) by ijuz\r\n","This was SIT",SIT_VERSION);
		}

		/* that is the lucky end... ;) */
		exit(0);
		/* now we are ended... or we've got a lot of shitt on screen
			if we use a great interface... this have to be examined
			FIXME !
			bye
		*/
	}

	/* HELP */
	else if (strcmp(cmd1,"help")==0)
	{
		/* some text... should get better */
		SC_help("ALL");
	}

	/* HELP */
	else if (strcmp(cmd1,"version")==0)
	{
		printf("%s\n",SIT_VERSION);
	}

	/* MSG */
	else if (strcmp(cmd1,"msg")==0)
	{
		if(cmd2!=NULL && input2!=NULL)
		{
			SIT_printf_to_tcpSend("PRIVMSG %s :%s\r\n",cmd2,input2);
		}
		else
		{
			printf("an error occured! \n  \\msg <nick> <message> is the right syntax");
			problem+=1;
		}
	}

	/* ME */
	else if (strcmp(cmd1,"me")==0)
	{
		SIT_printf_to_tcpSend("PRIVMSG %s :\001ACTION %s\001\r\n",DS_s->PCHANNEL,input1);
	}

	/* NICK */
	else if(strcmp(cmd1,"nick")==0)
	{
		if(cmd2!=NULL)
		{
			SIT_printf_to_tcpSend("NICK %s\r\n",cmd2);
		}
		else
		{
			printf("an error occured! \n  \\nick <new-nick> is the right syntax");
			problem+=1;
		}
	}

	/* JOIN */
	/*because SIT is not able to join more then 1 channel at once, the
	old is first PARTed */
	else if(strcmp(cmd1,"join")==0)
	{
		if(cmd2!=NULL)
		{
			SIT_printf_to_tcpSend("PART %s\r\n",DS_s->PCHANNEL);
			strcpy(DS_s->PCHANNEL,cmd2);
			SIT_printf_to_tcpSend("JOIN %s\r\n",DS_s->PCHANNEL);
		}
		else
		{
			printf("an error occured! \n  \\join <new-channel> is the right syntax");
			problem+=1;
		}
	}

	/* QUOTE */
	else if (strcmp(cmd1,"quote")==0)
	{
		if(input1!=NULL)
		{
			SIT_printf_to_tcpSend("%s\r\n",input1);
		}
		else
		{
			printf("an error occured! \n  \\quote <to-send> is the right syntax");
			problem+=1;
		}
	}

	/* PING */

	else if (strcmp(cmd1,"ping")==0)
	{
		if(cmd2!=NULL)
		{
			SIT_printf_to_tcpSend("PRIVMSG %s :\001PING %ld\001\r\n",cmd2,time(NULL));
		}
		else
		{
			printf("an error occured! \n  /ping <nick> is the right syntax\n");
		}
	}

	else if (strcmp(cmd1,"db")==0)
	{
		if(cmd2!=NULL)
		{
			if(strcmp(cmd2,"help")==0)
			{
				printf("%s\n","#### Help for the /db command ####");
				printf("%s\n","this command is for debuging");
				printf("%s\n","/db <command>");
				printf("%s\n","ganick - put out all nicks in the datastructure");
				printf("%s\n","gachannel - put out all nicks in the datastructure");
				
			}
			else if(strcmp(cmd2,"ganick")==0)
			{
				DS_nick_giveall(DS_s);
			}
			else if(strcmp(cmd2,"gachannel")==0)
			{
				DS_channel_giveall(DS_s);
			}
			else
			{
				printf("%s\n","ERROR: try /db help for possible commands");
			}
		}
		else
		{
			printf("%s\n","ERROR: try /db help for possible commands");
		}
	}
		
	else
	{
		printf("unknown or no command detected\n");
		printf("please make /help\n");
		problem+=1;
	}
	
	/* now lets clean up */
	free(cmd1);
	if(cmd2!=NULL){free(cmd2);}
	if(problem==0){return 1;}
		else{return 0;}
}


int main(int argc, char *argv[])
{
#ifdef DEBUG
	FILE *fp;
#endif
	int i;
	Cmdline *cmd; /* Cmdline is a structure for the possible arguments */ 

	/* Switch the buffer size for stdin and stdout to 1 line
		this is important for using a overlaying GUI */
	setvbuf(stdout, NULL, _IOLBF, 0);

#ifdef __EMX__
	tcgetattr(fileno(stdin), &tio);
	tio.c_lflag &= ~IDEFAULT;
	tcsetattr(fileno(stdin), TCSANOW, &tio);
#endif
	/* now the structure for the arguments will be fulled */
	cmd = parseCmdline(argc, argv);

	/* first we check if an option for help was given */
	if(cmd->helpP!=0){usage();}
	if(cmd->helplongP!=0){usage();}

	/* then if the version was requested */
	if(cmd->versionP!=0)
	{
		printf("%s %s\n","This is SIT Version",SIT_VERSION);
		exit(0);
	}

	/* now lets initializise the server data structure
	   at this state we do only use a single structure, a list is a this
	   state unneccessary, but this should be extensible */
	DS_s = (struct node_s_list *)malloc(sizeof(*DS_s));
	DS_s->DS_channel=NULL;
	for(i=0;i<DS_NICK_HASHSIZE;i++) DS_s->DS_nick[i]=NULL;
	DS_s->NICKNAME = NULL;
	DS_s->SERVER = NULL;
	DS_s->SOCKET = 0; /* int */
	DS_s->PCHANNEL = NULL;
	DS_s->SOCKBUF_SIZE=CONFIG_SIT_SOCKBUF_SIZE;
	DS_s->SOCKBUF_POS=0;
	DS_s->SOCKBUF = (char *)malloc(DS_s->SOCKBUF_SIZE);
	DS_s->SOCKBUF[0] = 0;

	if(cmd->nicknameP!=0) DS_s->NICKNAME = strdup(cmd->nickname);
	else
#if defined __CYGWIN__
	{
		printf("ERROR: because it's very probable that you've got no USER environment variable\n");
		printf("BYE!\n");
		exit(1);
	}
#else
		 DS_s->NICKNAME = strdup(getenv("USER"));
#endif

	if(cmd->serverP!=0) DS_s->SERVER = strdup(cmd->server);
		else DS_s->SERVER = strdup(DEFAULT_SERVER);

	if(cmd->channelP!=0){DS_s->PCHANNEL = strdup(cmd->channel);}
		else {DS_s->PCHANNEL = strdup(DEFAULT_CHANNEL);}

	/* FIXME */
	BLOWFISH_TEMP = DS_blowfish_init();
	
	printf("%s %s%s\n","SIT",SIT_VERSION,", Copyright (C) 2000 Christian Leber");
	printf("%s\n","SIT comes with ABSOLUTELY NO WARRANTY; for details see the file copying.");
	printf("%s\n","This is free software, and you are welcome to redistribute it");
	printf("%s\n","under certain conditions; see the file copying for details.");
	printf("\n");
	printf("%s\n","-----------------------------------------");
	printf("%s %s\n","Server to connect:   ",DS_s->SERVER);
	printf("%s %s\n","Your nickname is:    ",DS_s->NICKNAME);
	printf("%s %s\n","Channel to join is:  ",DS_s->PCHANNEL);
	printf("%s\n","-----------------------------------------");

#ifdef DEBUG
	fp = fopen("log.sit","a");
	fprintf(fp,"######################## STARTING SIT #########################\n");
	fprintf(fp,"Connecting to: %s as: %s\n",DS_s->SERVER, DS_s->NICKNAME);
	fclose(fp);
#endif

	/* first we should connect to the server */
	if(!SIT_Connect(DS_s->SERVER,6667)) {printf("%s\n","Cannot connect");exit(1);}
	
	/* now we say the server witch nick we would like to have */
	SIT_printf_to_tcpSend("NICK %s\r\n",DS_s->NICKNAME);
	
	/* the server would like to know our user and realname... */
	SIT_printf_to_tcpSend("USER %s 0 0 :%s\r\n",DEFAULT_USERNAME,DEFAULT_REALNAME);

	/* it's a good time to go to select to look at stdin and the socket */
	SIT_select();
	exit(1);  
}

void SIT_select(void)
{
	fd_set rset;
	int maxfdp1;	
	char buffer[256] = "";

	FD_ZERO(&rset);
	
	for( ; ; )
	{
	FD_SET(fileno(stdin), &rset);
	FD_SET(DS_s->SOCKET, &rset);
	maxfdp1 = max(fileno(stdin),DS_s->SOCKET) + 1;

		select(maxfdp1,&rset,NULL,NULL,NULL);
		
		if(FD_ISSET(fileno(stdin),&rset))
		{
			fgets(buffer, 256, stdin);
			SIT_parse_input(buffer);
		}

		if(FD_ISSET(DS_s->SOCKET,&rset))
		{
			socket_to_buffer();	
			while(buffer_to_msg());
		}
	}
}


void SC_help(const char *cmd)
{
	if(strcmp(cmd,"ALL")==0)
	{
		printf("%s\n","--------------------------------------------------------------");
		printf("%s\n","Hello User, this is the Help text to SIT");
		printf("%s\n","SIT means SIS is tiny, but no more ;)");
		printf("%s %s\n\n","You are using",SIT_VERSION);
		printf("There are the following commands:\n");
		printf("/help - this helpfull text\n");
		printf("/msg <nick or channel> <message>\n");
		printf("/quit <quitmessage> to leave SIT\n");
		printf("/exit the same as quit SIT\n");
		printf("/nick <new-nick>\n");
		printf("/join <new channel to join>\n");
		printf("/me <what you do>\n");
		printf("/db <command> - type /db help\n");
		printf("/quote <send this to the server>\n");
		printf("%s\n","--------------------------------------------------------------");
	}
}


void socket_to_buffer(void)
{
	int length;
	
	if((DS_s->SOCKBUF_SIZE-DS_s->SOCKBUF_POS) >= 0)
	{
		length = recv(DS_s->SOCKET,DS_s->SOCKBUF+DS_s->SOCKBUF_POS,
				DS_s->SOCKBUF_SIZE-DS_s->SOCKBUF_POS-1,0);
		if( length == -1) {printf("shitt happens, in socket_to_buffer\n");return;}
		else
		{
			DS_s->SOCKBUF[DS_s->SOCKBUF_POS+length] = 0;
			DS_s->SOCKBUF_POS += length;
		}
    }
	else
	{
		printf("der liest nicht \n");
		sleep(3);
	}
}


bool buffer_to_msg(void)
{
	int len;
	char *buf,*pos;
	
	FILE *fp;

	pos = strstr(DS_s->SOCKBUF,"\r\n");
	if(pos != NULL)
	{
		len = pos - DS_s->SOCKBUF;
		buf = (char *)malloc(len+1);
		memcpy(buf,DS_s->SOCKBUF,len);
		buf[len] = 0;
 		memmove(DS_s->SOCKBUF,pos+2,DS_s->SOCKBUF_POS-len-2);
		DS_s->SOCKBUF_POS -= (len+2);
		DS_s->SOCKBUF[DS_s->SOCKBUF_POS] = 0;
#ifdef DEBUG
		fp = fopen("log.sit","a");
		fprintf(fp,"%s\n",buf);
		fclose(fp);
		
		fp = fopen("log.server.raw","a");
		fprintf(fp,"%s\n",buf);
		fclose(fp);
		
#endif
		if(strlen(buf) > 512)
		{
			printf("ERROR: buffer_to_msg: message to long (>512)\n");
			fp = fopen("ERROR.sit","a");
			fprintf(fp,"ERROR: buffer_to_msg: message to long (>512)\n"); 
			fprintf(fp,"   {%s}\n",buf);
			fclose(fp);
			free(buf);
			return false;
		}
				
		if(!fserver_in_minimal(buf))
		{
			if(!fserver_in(buf))
			{
				if(!fserver_in2(buf))
				{
					printf("### THERE IS A MESSAGE FROM THE SERVER I CAN NOT INTERPRET: ###\n");
					printf("%s\n",buf);
					printf("##### please send your log.left file to the maintainer ########\n");
					fp = fopen("log.left","a");
					fprintf(fp,"%s\n",buf);
					fclose(fp);
				}
			}
		}
		free(buf);
		return true;
	}
	return false;
}


bool SIT_tcpSend(char *input)
{
	int len; 
#ifdef DEBUG
	FILE *fp;
  	fp = fopen("log.sent","a");
	fprintf(fp,"%s|\n",input);
	fclose(fp);
#endif

	len = strlen(input);
	if(send(DS_s->SOCKET,input,len,0)!=len)
	{
		printf("%s\n","ERROR: SIT_tcpSend, the number\
		 of bytes to send and the return value of send are NOT the same");	
		return false;
	}
	return true;
}

void SIT_printf_to_tcpSend(char *format, ...)
{
	char *buff1; /* general buffer... you know that :)) */
	int retval=0; /* return value from snprintf */
	va_list arg_list;
	
	buff1 = (char *)malloc(512);

	va_start(arg_list, format);
	retval=vsnprintf(buff1,1024,format,arg_list);
	va_end(arg_list);
	
	if(retval==-1 || retval>=511)
	{
		printf("%s\n", "ERROR: shitt happens in SIT_printf_to_tcpSend;\
		 	to much stuff for the buffer");
		return;
	}
	SIT_tcpSend(buff1);
}

